from container_hub.config import settings
from container_hub.carriers import container_terminal

up = container_terminal.up
down = container_terminal.down
container_list = container_terminal.container_list



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
import logging
from pathlib import Path
from typing import Dict, List, Optional
from typing import Iterator
from typing import Tuple
import marathon.exceptions
from marathon import MarathonApp
from marathon.models.app import MarathonTask
from marathon.models.container import MarathonContainer
from marathon.models.container import MarathonContainerVolume
from marathon.models.container import MarathonDockerContainer
from marathon.models.constraint import MarathonConstraint
from marathon import MarathonClient

from container_hub.exceptions import CarrierError
from container_hub import settings

logger = logging.getLogger(__name__)

client = MarathonClient(settings.CLIENT_URL)


class MachineManagerException(Exception):
    pass


def container_list() -> List[str]:
    sim_uids = []
    apps = client.list_apps()
    for app in apps:
        # v3 apps all have a simulation id label
        sim_uid = app.labels.get("simulation_id")
        if not sim_uid:
            continue
        sim_uids.append(sim_uid)
    return sim_uids


def _task_generator() -> Iterator[Tuple[MarathonTask, str]]:
    tasks = client.list_tasks()
    for task in tasks:
        app_name = task.app_id.strip("/")
        yield task, app_name


def container_hosts() -> Dict[str, str]:
    d = {}
    for task, app_name in _task_generator():
        d[app_name] = task.host
    return d


def container_ips() -> Dict[str, str]:
    d = {}
    for task, app_name in _task_generator():
        _ip = task.ip_addresses[0]
        d[app_name] = _ip.ip_address
    return d


def up(
    sim_uid: str,
    sim_ref_datetime: str,
    end_time: int,
    duration: int,
    pause_timeout: int,
    start_mode: str,
    model_config: str,
    session_memory: int,
    envs: Dict,
    labels: Dict,
    max_rate: float = 0.0,
    clean_up_files: bool = False,
    gridadmin_download_url: Optional[str] = None,
    tables_download_url: Optional[str] = None
) -> str:

    """Create a MarathonApp instance."""
    image = settings.IMAGE_NAME
    name = f"simulation-{sim_uid}"
    labels.update({"simulation_id": f"{sim_uid}"})

    docker_container = MarathonDockerContainer(
        image=image,
        network="BRIDGE",
        parameters=[{"key": "network", "value": settings.NETWORK_NAME}],
    )

    volumes = []  # == mounts

    skip_model_mount: bool = all([x for x in [gridadmin_download_url, tables_download_url]])

    for local_path, mount_setting in settings.DOCKER_MOUNT_POINTS.items():
        if skip_model_mount and mount_setting['bind'] == '/models':
            # Skip mounting models
            continue

        volumes.append(
            MarathonContainerVolume(
                container_path=mount_setting["bind"],
                host_path=local_path,
                mode="RO" if mount_setting["ro"] else "RW",
            )
        )

    logger.debug("Volumes %s", volumes)

    # docker container with volumes
    container = MarathonContainer(docker=docker_container, volumes=volumes)
    result_path = Path(settings.BASE_RESULT_PATH, name)
    # environment variables for container
    envs.update({"RESULT_PATH": result_path.as_posix()})

    # Sentry environment variables for container
    if hasattr(settings, "SENTRY_DSN"):
        envs.update(
            {
                "SENTRY_DSN": settings.SENTRY_DSN,
                "RELEASE": getattr(settings, "SENTRY_RELEASE", ""),
            }
        )

    if hasattr(settings, "CONTAINER_LOG_LEVEL"):
        envs.update({"LOG_LEVEL": settings.CONTAINER_LOG_LEVEL})

    constraints = []

    if hasattr(settings, "MARATHON_CONSTRAINTS"):
        constraints = [
            MarathonConstraint(param, operator, value)
            for param, operator, value in settings.MARATHON_CONSTRAINTS
        ]

    # all args must be strings
    args = [
        "python",
        "service.py",
        settings.REDIS_HOST,
        model_config,
        sim_uid,
        sim_ref_datetime,
        str(end_time),
        str(duration),
        start_mode,
        str(pause_timeout),
        str(max_rate),
        str(clean_up_files)
    ]

    if gridadmin_download_url is not None:
        args.append(str(gridadmin_download_url))
    if tables_download_url is not None:
        args.append(str(tables_download_url))

    marathon_app_definition = MarathonApp(
        args=args,
        container=container,
        mem=session_memory,
        cpus=settings.CONTAINER_CPUS,
        env=envs,
        labels=labels,
        constraints=constraints,
    )

    try:
        app = client.create_app(name, marathon_app_definition)
    except marathon.exceptions.MarathonHttpError as err:
        logger.exception("Failed to create app %s with error %s", name, err)
        raise CarrierError(err)
    logger.info(f"App {app.id} started")
    return app.id


def _del_app(app_name: str) -> bool:
    """
    wrapper around the ``client.delete_app`` call

    :returns True if the app has been deleted, False otherwise
    """
    try:
        client.delete_app(app_name, force=True)
    except marathon.exceptions.NotFoundError:
        # not found; assume that the app is deleted
        logger.exception("Failed to delete app %s.", app_name)
        return False
    return True


def down(sim_uid: str):
    """Remove the given app."""

    app_name = f"simulation-{sim_uid}"

    # should return a single marathon.models.app.MarathonApp instance
    apps = client.list_apps(app_id=app_name)
    if not apps:
        logger.warning(
            "App not found; assuming that the app has already been deleted"
        )
        return

    if len(apps) > 1:
        msg = (
            f"Found more than one app that matches the name {app_name}. "
            f"Trying to delete all of them now..."
        )
        logger.warning(msg)
        for app in apps:
            _del_app(app.id)
        return

    app = apps[0]
    if _del_app(app.id):
        logger.info("Deleted app %s.", app_name)



import pytest
import os


@pytest.fixture
def docker_simple_settings():
    # settings = LazySettings('tests.test_files.settings')
    os.environ.update({"SIMPLE_SETTINGS": "tests.test_files.docker_settings"})
    from container_hub.config import get_settings
    yield get_settings()
    os.environ.pop("SIMPLE_SETTINGS")

@pytest.fixture
def docker_django_settings():
    # settings = LazySettings('tests.test_files.settings')
    os.environ.update({"DJANGO_SETTINGS_MODULE": "tests.test_files.django_settings"})
    from container_hub.config import get_settings
    yield get_settings()
    os.environ.pop("DJANGO_SETTINGS_MODULE")


def test_import_error():
    with pytest.raises(ImportError):
        from container_hub import settings

def test_smoke_django(docker_django_settings):
    assert docker_django_settings.CONTAINER_CARRIER == 'docker'


def test_smoke_simple(docker_simple_settings):
    assert docker_simple_settings.CONTAINER_CARRIER == 'docker'





from container_hub import settings

import importlib

__all__ = ["container_terminal"]


def __getattr__(name):
    if name in __all__:
        carrier = settings.CONTAINER_CARRIER.lower()
        mod = f".{carrier}.{name}"
        return importlib.import_module(mod, __name__)

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



def get_settings():

    try:
        from simple_settings import settings
        _ = settings.CONTAINER_CARRIER
        return settings
    except (RuntimeError, ImportError, ModuleNotFoundError):
        pass
    try:
        from django.conf import settings
        from django.core.exceptions import ImproperlyConfigured
        _ = settings.CONTAINER_CARRIER
        return settings
    except (ImportError, ImproperlyConfigured):
        pass
    raise ImportError(
        'Could not load settings. First tried simple_settings, '
        'then django settings'
    )

settings = get_settings()



import os

DEBUG = True
CONTAINER_CARRIER = "docker"
CLIENT_URL = "unix://var/run/docker.sock"
NETWORK_NAME = os.environ.get("CALCCORE_NETWORK_NAME")  # "threedi_backend"
IMAGE_NAME = os.environ.get("CALCCORE_DOCKER_IMAGE")
CONTAINER_LOG_LEVEL = os.environ.get("CALCCORE_LOG_LEVEL", "INFO")

# harbor.lizard.net/threedi/threedicore:1.4.15.devtest_3"
REDIS_HOST = os.environ.get("REDIS_HOST", "redis")  # "redis"
BASE_MODEL_PATH = os.environ.get("CALCORE_MODEL_PATH", "/models")
# # Location where to write the results of the calccore.
BASE_RESULT_PATH = os.environ.get("CALCCORE_RESULTS_PATH", "/results")

THREEDI_OUTPUT_DIR = ""

DOCKER_MOUNT_POINTS = {}

# Provision docker mount points from all CALCCORE_MOUNT_XXXX
# environment variables
# format:
#    CALCCORE_MOUNT_XX=LOCAL_PATH:DOCKER_PATH
for key in os.environ.keys():
    if key.startswith('CALCCORE_MOUNT'):
        local_path, docker_path, read_only = os.environ[key].split(':')
        DOCKER_MOUNT_POINTS.update(
            {
                local_path: {
                    'bind': docker_path,
                    'ro': read_only == 'ro'}
            }
        )


START_ACTION_SUBCHANNEL_NAME = "start_action"
SENTRY_DSN = ""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
import codecs
import re
import os


here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'marathon>=0.11.0',
    'docker>=3.7.2,<4.3',
]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Lars Claussen",
    author_email='claussen.lars@nelen-schuurmans.nl',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="The 3Di simulation container hub",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='container_hub',
    name='container_hub',
    packages=find_packages(include=['container_hub', 'container_hub.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/nens/container-hub',
    version=find_version("container_hub", "version.py"),
    zip_safe=False,
)



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
# -*- coding: utf-8 -*-
import logging
from typing import List, Dict, Optional
from pathlib import Path

from docker import DockerClient
from docker.models.containers import Container
from docker.errors import DockerException
from docker.errors import APIError
from docker.errors import ContainerError
from docker.errors import NotFound
from docker.types import Mount

from container_hub.exceptions import CarrierError
from container_hub import settings

logger = logging.getLogger(__name__)

client = DockerClient(base_url=settings.CLIENT_URL)


def _list_containers(filters=None) -> List[Container]:
    return client.containers.list(filters=filters)


def container_list() -> List[str]:
    """
    Returns a list of simulation_ids
    """
    lc = _list_containers(filters={"label": "simulation_id"})
    return [x.name.lstrip("simulation-") for x in lc]


def container_hosts() -> Dict:
    return {}


def container_ips() -> Dict[str, str]:
    d = {}
    containers = _list_containers(filters={"label": "simulation_id"})
    for container in containers:
        try:
            ip_address = container.attrs["NetworkSettings"]["Networks"][
                settings.NETWORK_NAME
            ]["IPAddress"]
        except KeyError:
            continue
        d[container.name] = ip_address
    return d


def up(
    sim_uid: str,
    sim_ref_datetime: str,
    end_time: int,
    duration: int,
    pause_timeout: int,
    start_mode: str,
    model_config: str,
    session_memory: int,
    envs: Dict,
    labels: Dict,
    max_rate: float = 0.0,
    clean_up_files: bool = False,
    gridadmin_download_url: Optional[str] = None,
    tables_download_url: Optional[str] = None
) -> str:

    """
    Create container based on simulation and threedimodel.

    :returns the container id
    """
    name = f"simulation-{sim_uid}"
    result_path = Path(settings.BASE_RESULT_PATH, name)
    labels.update({"simulation_id": f"{sim_uid}"})

    _envs = [f"{key}={value}" for key, value in envs.items()]
    _envs.append(f"RESULT_PATH={result_path.as_posix()}")

    # Sentry environment variables for container
    if hasattr(settings, "SENTRY_DSN"):
        release = getattr(settings, "SENTRY_RELEASE", "")
        _envs += [f"SENTRY_DSN={settings.SENTRY_DSN}", f"RELEASE={release}"]

    if hasattr(settings, "CONTAINER_LOG_LEVEL"):
        _envs += [f"LOG_LEVEL={settings.CONTAINER_LOG_LEVEL}"]

    cmd = f"python service.py {settings.REDIS_HOST} {model_config} {sim_uid} {sim_ref_datetime} {end_time} {duration} {start_mode} {pause_timeout} {max_rate} {clean_up_files}" # noqa


    skip_model_mount: bool = all([x for x in [gridadmin_download_url, tables_download_url]])

    if gridadmin_download_url is not None:
        cmd += f" {gridadmin_download_url}"
    if tables_download_url is not None:
        cmd += f" {tables_download_url}"

    logger.debug("cmd %s", cmd)
    logger.debug("Envs %s", _envs)

    mounts = []
    for local_path, mount_setting in settings.DOCKER_MOUNT_POINTS.items():
        if skip_model_mount and mount_setting['bind'] == '/models':
            # Skip mounting models
            continue
        mounts.append(
            Mount(
                mount_setting["bind"],
                local_path,
                type="bind",
                read_only=mount_setting["ro"],
            )
        )

    client = DockerClient(base_url=settings.CLIENT_URL)

    try:
        container = client.containers.run(
            image=settings.IMAGE_NAME,
            command=cmd,
            name=name,
            network=settings.NETWORK_NAME,
            mounts=mounts,
            environment=_envs,
            detach=True,
            labels=labels,
        )
    except (DockerException, APIError, ContainerError) as err:
        logger.error(err)
        raise CarrierError(err)

    # double check everything went right
    try:
        client.containers.get(container.id)
    except (APIError, NotFound) as err:
        logger.error(
            f"simulation container exited prematurely. Could not retrieve "
            f"the container if though it should be running {err}"
        )
        raise CarrierError(err)

    logger.info(f"Started simulation container {container.name}")
    return container.id


def down(sim_uid: str):

    try:
        container = client.containers.get(f"simulation-{sim_uid}")
        container_id = container.id
    except (APIError, NotFound) as err:
        logger.error(
            f"Could not get the simulation container, error message: {err}"
        )
        raise CarrierError(err)
    try:
        if settings.DEBUG is False:
            container.remove(force=True)
        else:
            container.kill()
    except APIError as err:
        logger.error(
            f"Could not kill/remove the "
            f"simulation container, error message: {err}"
        )
        raise CarrierError(err)
    logger.info(f"Removed container for simulation {sim_uid}")
    return container_id
